/* console.c -- debug console
   Copyright (C) 1998 Jerome Thoen

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  */

#include "config.h"

#include <stdio.h>
#include <ctype.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <unistd.h>

#include "emuto8.h"
#include "emu6809.h"
#include "display.h"
#include "memory.h"
#include "console.h"
#include "video.h"
#include "disk.h"
#include "audio.h"
#include "int6809.h"
#include "version.h"

static char initcardfile[256];
static char romdirpath[256];

int activate_console = 0;
static int console_active = 0;

int debugio = 0;
int audio = 0;

static void sigbrkhandler(int sigtype)
{
  if (!console_active)
    activate_console = 1;
  signal(SIGINT, sigbrkhandler);
}

static void setup_brkhandler(void)
{
  signal(SIGINT, sigbrkhandler);
}

void console_init(void)
{
  printf("EmuTO8 " __EMUVERSION__ " beta - Thomson TO8 Emulator\nCopyright (c) 1998 Jerome Thoen\n\n");
}

void execute()
{
  int n;

  if (quickvideo)
    do {
      n = m6809_execute();
      video_do_cycle_fast(n);
    } while (n > 0 && !activate_console);
  else
    do {
      n = m6809_execute();
      video_do_cycle(n);
    } while (n > 0 && !activate_console);

  if (n < 0)
    printf("m6809 emulator return code %d\n", n);
}

void execute_addr(tt_u16 addr)
{
  int n = 0;

  if (rpc != addr)
    if (quickvideo)
      do {
	n = m6809_execute();
	video_do_cycle_fast(n);
      } while (n > 0 && !activate_console && rpc != addr);
    else
      do {
	n = m6809_execute();
	video_do_cycle(n);
      } while (n > 0 && !activate_console && rpc != addr);
    
  if (n < 0)
    printf("m6809 emulator return code %d\n", n);
}

void ignore_ws(char **c)
{
  while (**c && isspace(**c))
    (*c)++;
}

tt_u16 readhex(char **c)
{
  tt_u16 val = 0;
  char nc;

  ignore_ws(c);

  while (isxdigit(nc = **c)){
    (*c)++;
    val *= 16;
    nc = toupper(nc);
    if (isdigit(nc)) {
      val += nc - '0';
    } else {
      val += nc - 'A' + 10;
    }
  }
  
  return val;
}

char *readstr(char **c)
{
  static char buffer[256];
  int i = 0;

  ignore_ws(c);

  while (!isspace(**c) && **c && i < 255)
    buffer[i++] = *(*c)++;

  buffer[i] = '\0';

  return buffer;
}

int readint(char **c)
{
  int val = 0;
  char nc;

  ignore_ws(c);

  while (isdigit(nc = **c)){
    (*c)++;
    val *= 10;
    val += nc - '0';
  }
  
  return val;
}

int more_params(char **c)
{
  ignore_ws(c);
  return (**c) != 0;
}

char next_char(char **c)
{
  ignore_ws(c);
  return *(*c)++;
} 
  
void console_command(int go)
{
  static char input[80], copy[80];
  char *strptr;
  tt_u16 memadr = 0, start, end;
  long n;
  int i, r;

  strcpy(copy, "");

  for(;;) {
    activate_console = 0;
    console_active = 1;
    if (go) {
      go = 0;
      strcpy(input, "z");
      strptr = input;
    } else {
      printf("> ");
      fflush(stdout);
      if(fgets(input, 80, stdin) == 0)
	return;
      if (strlen(input) == 1)
	strptr = copy;
      else {
	strptr = strcpy(copy, input);
	if (*strptr >= 'A' && *strptr <= 'Z')
	  *strptr = *strptr - 'A' + 'a';
      }
    }

    switch (next_char(&strptr)) {
#ifdef MEM_TRAP
    case 'a' :
      if (more_params(&strptr)) {
	memtrapaddr = readhex(&strptr);
	memtrap = 1;
      } else
	printf("Syntax Error. Type 'h' to show help.\n");
      break;
#endif
    case 'b' :
      if (more_params(&strptr)) {
	char *filename = readstr(&strptr);
	
	if (memory_save(filename))
	  printf("Memory saved\n");
	else
	  printf("Error, memory not saved. Does directory \"sav/\" exist ?\n");
      } else
	printf("Syntax Error. Type 'h' to show help.\n");
      break;
    case 'c' :
      if (more_params(&strptr)) {
	char *filename = readstr(&strptr);

	if (memory_load(filename))
	  printf("Memory loaded\n");
	else
	  printf("Error, memory not loaded\n");
      } else
	printf("Syntax Error. Type 'h' to show help.\n");
      break;
    case 'd' :
      if (more_params(&strptr)) {
	start = readhex(&strptr);
	if (more_params(&strptr))
	  end = readhex(&strptr);
	else
	  end = start;
      } else 
	start = end = memadr;
      
      for(n = start; n <= end && n < 0x10000; n += dis6809((tt_u16)n, stdout));

      memadr = (tt_u16)n;
      break;
    case 'f' :
      if (more_params(&strptr)) {
	console_active = 0;
	execute_addr(readhex(&strptr));
	m6809_dumpregs();
	printf("Next PC:");
	memadr = rpc + dis6809(rpc, stdout);
      } else
	printf("Syntax Error. Type 'h' to show help.\n");
      break;
    case 'g' :
      if (more_params(&strptr))
	rpc = readhex(&strptr);
      console_active = 0;
      display_raise_window();
      execute();
      m6809_dumpregs();
      printf("Next PC:");
      memadr = rpc + dis6809(rpc, stdout);
      break;
    case 'h' : case '?' :
      printf("     HELP for the m6809 emulator debugger\n\n");
#ifdef MEM_TRAP
      printf("   a adr           : step forward until memory access at <adr>\n");
#endif
      printf("   b name          : save TO8 memory\n");
      printf("   c name          : load TO8 memory\n");
      printf("   d [start] [end] : disassemble memory from <start> to <end>\n");
      printf("   f adr           : step forward until PC = <adr>\n");
      printf("   g [adr]         : start execution at current address or <adr>\n");
      printf("   h, ?            : show this help page\n");
      printf("   i               : toggle debug IO\n");
      printf("   l name [side]   : load SAP disk file <name> on side <side>\n");
      printf("   m [start] [end] : dump memory from <start> to <end>\n");
      printf("   n [n]           : next instruction\n");
#ifdef SOUND
      printf("   o               : toggle audio and sync\n");
#endif
      printf("   p adr           : set PC to <adr>\n");
      printf("   q               : quit the emulator\n");
      printf("   r               : dump CPU registers\n");
#ifdef PC_HISTORY
      printf("   s               : show PC history\n");
      printf("   t               : flush PC history\n");
#endif
      printf("   u               : dump custom chips state\n");
      printf("   v               : toggle fast video\n");
      printf("   y               : hard resert\n");
      printf("   z               : reset CPU\n");
      break;
    case 'i' :
      debugio ^= 1;
      printf("Debug IO %s\n", debugio ? "active" : "inactive");
      break;
    case 'l' :
      if (more_params(&strptr)) {
	char *filename = readstr(&strptr);

	if (more_params(&strptr))
	  n = readint(&strptr) & 1;
	else
	  n = 0;

	if(!LoadSAP(filename, n))
	  printf("Error, SAP disk file %s not loaded\n", filename);
      } else
	printf("Syntax Error. Type 'h' to show help.\n");
      break;
    case 'm' :
      if (more_params(&strptr)) {
	n = readhex(&strptr);
	if (more_params(&strptr))
	  end = readhex(&strptr);
	else
	  end = n;
      } else 
	n = end = memadr;
      while (n <= (long)end) {
	printf("%04hX: ", (unsigned int)n);
	for (i = 1; i <= 8; i++)
	  printf("%02X ", get_memb(n++));
	n -= 8;
	for (i = 1; i <= 8; i++) {
	  tt_u8 v;

	  v = get_memb(n++);
	  if (v >= 0x20 && v <= 0x7e)
	    putchar(v);
	  else
	    putchar('.');
	}
	putchar('\n');
      }
      memadr = n;
      break;
    case 'n' :
      if (more_params(&strptr))
	i = readint(&strptr);
      else
	i = 1;

      while (i-- > 0) {
	r = m6809_execute();
	if (quickvideo)
	  video_do_cycle_fast(r);
	else
	  video_do_cycle(r);
	if (r <= 0) {
	  printf("m6809 emulator return code %d\n", r);
	  break;
	}
      }
      
      if (quickvideo)
	video_refresh();
      display_update();
      printf("Next PC:");
      memadr = rpc + dis6809(rpc, stdout);
      break;
#ifdef SOUND
    case 'o' :
      if (more_params(&strptr))
	printf("Syntax Error. Type 'h' to show help.\n");
      else
	if (audio) {
	  audio_quit();
	  audio = 0;
	  printf("Sound and sync desactivated\n");
	} else
	  audio = audio_init();
      break;
#endif
    case 'p' :
      if(more_params(&strptr))
	rpc = readhex(&strptr);
      else
	printf("Syntax Error. Type 'h' to show help.\n");
      break;
    case 'q' :
      return;
    case 'r' :
      m6809_dumpregs();
      break;
#ifdef PC_HISTORY
    case 's' :
      r = pchistidx - pchistnbr;
      if (r < 0)
	r += PC_HISTORY_SIZE;
      for (i = 1; i <= pchistnbr; i++) {
	dis6809(pchist[r++], stdout);
	if (r == PC_HISTORY_SIZE)
	  r = 0;
      }
      break;
    case 't' :
      pchistnbr = pchistidx = 0;
      break;
#endif
    case 'u' :
      mem_dumpstate();
      break;
    case 'v' :
      quickvideo ^= 1;
      printf("%s video\n", quickvideo ? "Fast" : "Slow but true");
      break;
    case 'y' :
      set_memw(0x60fe, 0);
      reset();
      console_active = 0;
      execute();
      m6809_dumpregs();
      memadr = rpc + dis6809(rpc, stdout);
      break;
    case 'z' :
      reset();
      console_active = 0;
      execute();
      m6809_dumpregs();
      memadr = rpc + dis6809(rpc, stdout);
      break;
    default :
      printf("Undefined command. Type 'h' to show help.\n");
      break;
    }
  }
}

void print_help(void)
{
  printf("EmuTO8 " __EMUVERSION__ " beta\nCopyright (c) 1998 Jerome Thoen\n");
  printf("Command line options :\n\n");
  printf("  -[0-1] file : load SAP disk file on side [0-1]\n");
  printf("  -b size     : screen border size (0:none,1:small,2:medium,3:full,4:640x480)\n");
  printf("  -c file     : use file as ROM card\n");
  printf("  -e          : extended memory (512kb instead of 256kb)\n");
  printf("  -h          : print help\n");
  printf("  -r dir      : specify system roms directory path\n");
  printf("  -x          : use X Shm extension\n");
}

void default_prefs(void)
{
  ramext = 0;
  initcardfile[0] = '\0';
  strcpy(romdirpath, "rom/");
}

void parse_cmdline(int argc, char **argv)
{
  int c;
  char a;

  while ((c = getopt(argc, argv, "0:1:b:c:dehr:x")) != EOF)
    switch (c) {
    case '0' :
	if(!(LoadSAP(optarg, 0)))
	  fprintf(stderr, "Error, SAP disk file %s not loaded\n", optarg);
	break;
    case '1' :
	if(!(LoadSAP(optarg, 1)))
	  fprintf(stderr, "Error, SAP disk file %s not loaded\n", optarg);
	break;
    case 'b' :
      a = optarg[0];
      switch (a) {
      case '0' :
	visible_cols = visible_lines = 0;
	break;
      case '1' :
	visible_cols = 2;
	visible_lines = 16;
	break;
      case '2' :
	visible_cols = 4;
	visible_lines = 32;
	break;
      case '3' :
	visible_cols = 12;
	visible_lines = 56;
	break;
      case '4' :
	visible_cols = 0;
	visible_lines = 20;
	break;
      default :
	fprintf(stderr, "Invalid number %s for border size\n", optarg);
	print_help();
	exit(0);
      }
      break;
    case 'c' :
      strncpy(initcardfile, optarg, 254);
      initcardfile[255] = '\0';
      break;
    case 'e' :
      ramext = 1;
      break;
    case 'h' :
    case '?' :
      print_help();
      exit(0);
    case 'r' :
      strncpy(romdirpath, optarg, 255);
      if (romdirpath[strlen(romdirpath) - 1] != '/')
	strcat(romdirpath, "/");
      romdirpath[255] = '\0';
      break;
    case 'x' :
#ifdef SHM_SUPPORT
      use_shm = 1;
#else
      printf("Shm support disabled at compilation time\n");
#endif
      break;
    }
  if (optind < argc) {
    fprintf(stderr, "Invalid option %s\n", argv[optind]);
    print_help();
    exit(0);
  }
}

int main(int argc, char **argv)
{
  default_prefs();
  parse_cmdline(argc, argv);
  console_init();
  m6809_init();
  if (!memory_init(romdirpath, initcardfile))
    return 1;
  install_trap();
  if (!display_init(NULL))
    return 1;
  video_init();
#ifdef SOUND
  audio = audio_init();
#endif
  setup_brkhandler();
  console_command(1);

#ifdef SOUND
  if (audio)
    audio_quit();
#endif

  return 0;
}
